#include <windows.h>
#include <stdio.h>

#include "bm.h"

#define MAX(a,b) (((a) > (b)) ? (a) : (b))




void preBmBc(unsigned char *x, int m, int bmBc[]) {
   int i;
 
   for (i = 0; i < ASIZE; ++i)
      bmBc[i] = m;
   for (i = 0; i < m - 1; ++i)
     bmBc[x[i]] = m - i - 1;
}
 
 
void suffixes(char *x, int m, int *suff) {
   int f, g, i;
 
   suff[m - 1] = m;
   g = m - 1;
   for (i = m - 2; i >= 0; --i) {
      if (i > g && suff[i + m - 1 - f] < i - g)
         suff[i] = suff[i + m - 1 - f];
      else {
         if (i < g)
            g = i;
         f = i;
         while (g >= 0 && x[g] == x[g + m - 1 - f])
            --g;
         suff[i] = f - g;
      }
   }
}
 
void preBmGs(char *x, int m, int bmGs[]) {
   int i, j, suff[m];
 
   suffixes(x, m, suff);
 
   for (i = 0; i < m; ++i)
      bmGs[i] = m;
   j = 0;
   for (i = m - 1; i >= 0; --i)
      if (suff[i] == i + 1)
         for (; j < m - 1 - i; ++j)
            if (bmGs[j] == m)
               bmGs[j] = m - 1 - i;
   for (i = 0; i <= m - 2; ++i)
      bmGs[m - 1 - suff[i]] = m - 1 - i;
}


bm_info* bm_init(char* pattern, size_t pattern_len) {

  bm_info* bmi;

  if (!pattern || !pattern_len) return NULL;


  // reserve memory
  bmi = malloc(sizeof(bm_info));
  
  if (!bmi) {
    perror("malloc() failed\n");
    return NULL;
  }

  bmi->skipG = calloc(sizeof(int), pattern_len);

  if (!bmi->skipG) {
    perror("calloc() failed\n");
    free(bmi);
    return NULL;
  }

  bmi->needle = calloc(sizeof(char), pattern_len);
  
  if (!bmi->needle) {
    perror("calloc() failed\n");
    free(bmi->skipG);
    free(bmi);
    return NULL;
  }

  memcpy(bmi->needle, pattern, pattern_len);
  bmi->needle_len = pattern_len;

  //create tables
  /* Preprocessing */
  preBmGs(pattern, pattern_len, bmi->skipG);
  preBmBc(pattern, pattern_len, bmi->skipB);

  return bmi;
}


void bm_free(bm_info* bmi){
  
  if (!bmi) return;

  if (bmi->needle) free(bmi->needle);
  if (bmi->skipG) free(bmi->skipG);
  free(bmi);
}
 
/* 
   boyer moore string search 
   return: memory address of matching char or NULL
*/ 
char* BM(unsigned char *haystack, int hay_len, bm_info* bmi) {
  register int i;
  register int j;

  if (!haystack || !bmi) return NULL;

  unsigned char* needle = bmi->needle;
  size_t needle_len = bmi->needle_len;
  size_t max_len = hay_len - needle_len;

  int* bmGs = bmi->skipG;
  int* bmBc = bmi->skipB;

   /* Searching */
   j = 0;
   while (j <= max_len) {     
     for (i = needle_len - 1; i >= 0 && needle[i] == haystack[i + j]; --i);
     if (i>=0) {
       j += MAX((int)bmGs[i], (int)(bmBc[haystack[i + j]] - needle_len + 1 + i));     
     }
     else
       return &haystack[j];                     
     
   }
   return NULL;
}
   

